from neuron import h, gui


class Model:
    def __init__(self):
        nlayer = int(h.nlayer_extracellular())

        cable = [h.Section("cable%d" % i) for i in range(4)]
        cable[1].connect(cable[0](1))
        cable[2].connect(cable[1](1))
        cable[3].connect(cable[0](0))

        for sec in cable:
            sec.nseg = 3
            sec.L = 100
            sec.diam = 1
            sec.insert("hh")
            sec.insert("extracellular")  # default, virtually wire to ground
            for i in range(nlayer):
                sec.xg[i] = 1e7  # Gaussian elimination roundoff errors in 10 ms
                sec.xraxial[i] = 1e14  # accumulate if left at default 1e9
            sec.xg[nlayer - 1] = 0.001  # all voltage drop in last layer

        cable[3](1).e_extracellular = -20  # stimulate extracellularly

        self.cable = cable

    def run(self):
        h.tstop = 10
        h.run()
        res = [h.Vector(), h.Vector()]
        for sec in self.cable:
            for seg in sec.allseg():
                res[0].append(seg.v)
                res[1].append(seg.vext[0])
        return res


def test_nlayer():
    res = []
    for nlayer in [1, 2, 3]:
        h.nlayer_extracellular(nlayer)
        assert nlayer == int(h.nlayer_extracellular())
        m = Model()
        res.append(m.run())
        del m
    for r in res[1:]:
        for i in range(2):
            x = r[i].c().sub(res[0][i]).abs().sum()
            print(x)
            assert x < 0.01  # At 10ms. At 5ms would be 2e-5
    return res


if __name__ == "__main__":
    res = test_nlayer()
    for i in range(len(res[0][0])):
        for j in range(2):
            for k in range(3):
                print("%g" % res[k][j].x[i], end=" ")
        print()



from neuron import h, rxd

h.load_file("stdrun.hoc")

cell = h.Section(name="cell")
r = rxd.Region([cell])

Ca = rxd.Species(r, name="ca", charge=2, initial=60e-3)

reaction_decay = rxd.Reaction(Ca, 0 * Ca, 1e-8)
tvec = h.Vector().record(h._ref_t)
cavec = h.Vector().record(Ca.nodes[0]._ref_concentration)
h.finitialize(-70)
h.continuerun(10)

if __name__ == "__main__":
    from matplotlib import pyplot

    pyplot.plot(tvec, cavec)
    pyplot.show()



import pytest
from testutils import compare_data, tol


@pytest.fixture
def empty_regions_model(neuron_instance):
    """A model with an empty region"""

    h, rxd, data, save_path = neuron_instance
    sec = h.Section()
    sec.L = 100
    sec.diam = 1
    sec.nseg = 100
    cells = []
    cyt = rxd.Region(cells, name="cyt", nrn_region="i")
    mem = rxd.Region(cells, name="mem", geometry=rxd.membrane)
    ecs = rxd.Extracellular(-5, -5, -5, 5, 5, 5, dx=5)
    k = rxd.Species(
        [ecs, cyt],
        name="k",
        d=0,
        charge=1,
        initial=lambda nd: 140 if hasattr(nd, "x") else 3,
    )

    model = (sec, cyt, mem, ecs, k)
    yield (neuron_instance, model)


def test_empty_region_no_rate(empty_regions_model):
    """Test a rate that should do nothing with an empty region"""

    neuron_instance, model = empty_regions_model
    h, rxd, data, save_path = neuron_instance
    sec, cyt, mem, ecs, k = model
    kcyt = k[cyt]
    react = rxd.Rate(kcyt, -10 * kcyt)
    h.finitialize(-70)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_empty_region_ecs_rate(empty_regions_model):
    """Test a rate that should change the ECS with an empty region"""

    neuron_instance, model = empty_regions_model
    h, rxd, data, save_path = neuron_instance
    sec, cyt, mem, ecs, k = model
    react = rxd.Rate(k, -10 * k)
    h.finitialize(-70)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_empty_region_no_reaction(empty_regions_model):
    """Test a reaction that should change the ECS with an empty region"""

    neuron_instance, model = empty_regions_model
    h, rxd, data, save_path = neuron_instance
    sec, cyt, mem, ecs, k = model
    react = rxd.Reaction(2 * k, k, 10, regions=[cyt])
    h.finitialize(-70)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_empty_region_ecs_reaction(empty_regions_model):
    """Test a reaction that should change the ECS with an empty region"""

    neuron_instance, model = empty_regions_model
    h, rxd, data, save_path = neuron_instance
    sec, cyt, mem, ecs, k = model
    react = rxd.Reaction(2 * k, k, 10)
    h.finitialize(-70)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol


def test_empty_region_multicompartment_reaction(empty_regions_model):
    """Test a reaction that should change the ECS with an empty region"""

    neuron_instance, model = empty_regions_model
    h, rxd, data, save_path = neuron_instance
    sec, cyt, mem, ecs, k = model
    kcyt, kecs = k[cyt], k[ecs]
    react = rxd.MultiCompartmentReaction(
        kcyt, kecs, 1e5, 0, membrane=mem, membrane_flux=False
    )
    h.finitialize(-70)
    h.continuerun(10)
    if not save_path:
        max_err = compare_data(data)
        assert max_err < tol



"""
neuron.doc
===================

This module is used behind the scenes to generate docstrings of all HocObjects
of various types

To get general help on the neuron module try:

>>> import neuron
>>> help(neuron)

or in Jupyter

In []: import neuron
In []: neuron ?

From there, you can get help on the various objects in the hoc world:

In []: fom neuron import h
In []: v = h.Vector()
In []: ? v.to_python

a feature whose implementation is based on the neuron.doc module.

"""

import pydoc, sys, io, inspect
from neuron import h


header = """
NEURON+Python Online Help System
================================

"""


# override basic helper functionality to give proper help on HocObjects
class NRNPyHelper(pydoc.Helper):
    def __call__(self, request=None):
        if type(request) == type(h):
            pydoc.pager(header + request.__doc__)
        else:
            pydoc.Helper.__call__(self, request)


help = NRNPyHelper(sys.stdin, sys.stdout)


def doc_asstring(thing, title="Python Library Documentation: %s", forceload=0):
    """return text documentation as a string, given an object or a path to an object."""
    try:
        object, name = pydoc.resolve(thing, forceload)
        desc = pydoc.describe(object)
        module = inspect.getmodule(object)
        if name and "." in name:
            desc += " in " + name[: name.rfind(".")]
        elif module and module is not object:
            desc += " in module " + module.__name__
        if not (
            inspect.ismodule(object)
            or inspect.isclass(object)
            or inspect.isroutine(object)
            or inspect.isgetsetdescriptor(object)
            or inspect.ismemberdescriptor(object)
            or isinstance(object, property)
        ):
            # If the passed object is a piece of data or an instance,
            # document its available methods instead of its value.
            object = type(object)
            desc += " object"
        return title % desc + "\n\n" + pydoc.text.document(object, name)
    except (ImportError, ErrorDuringImport) as value:
        print(value)


# override systemwide behaviour
pydoc.help = help


doc_h = """

neuron.h
========

neuron.h is the top-level HocObject, allowing interaction between Python and Hoc.

It is callable like a function, and takes Hoc code as an argument to be executed.

The top-level Hoc namespace is exposed as attributes to the h object.

Ex:

    >>> h = neuron.h
    >>> h("objref myvec")
    >>> h("myvec = new Vector(10)")
    >>> h.myvec.x[0]=1.0
    >>> h.myvec.printf()
    1       0       0       0       0
    0       0       0       0       0


NEURON classes are defined in the h namespace and can be constructed as follows:

>>> v = h.Vector(10)
>>> soma = h.Section()
>>> input = h.IClamp(soma(0.5))

More help on individual classes defined in Hoc and exposed in Python
is available using Jupyter's online help feature

In []: ? h.Section

or in standard Python by python help system

>>> help(h.Vector)

For a list of symbols defined in neuron.h try:

>>> dir(neuron.h)



NOTE: Several Hoc symbols are not useful in Python, and thus raise an exception when accessed, for example:

In []: h.objref
---------------------------------------------------------------------------
TypeError                                 Traceback (most recent call last)

/home/emuller/hg/nrn_neurens_hg/<ipython console> in <module>()

TypeError: Cannot access objref (NEURON type 323) directly.

In []: h.objref ?
Object `h.objref` not found.


"""


default_class_doc_template = """
No docstring available for class '%s'

Try checking the online documentation at:
https://www.neuron.yale.edu/neuron/static/py_doc/index.html
"""


default_object_doc_template = """
No docstring available for object type '%s'

Try checking the online documentation at:
https://www.neuron.yale.edu/neuron/static/py_doc/index.html
"""


default_member_doc_template = """
No docstring available for the class member '%s.%s'

Try checking the online documentation at:
https://www.neuron.yale.edu/neuron/static/py_doc/index.html

==================================================

%s
    
"""


def _get_from_help_dict(name):
    return _help_dict.get(name, "")


def _get_class_from_help_dict(name):
    result = _get_from_help_dict(name)
    if not result:
        return ""
    methods = dir(h.__getattribute__(name))
    for m in methods:
        if name + "." + m in _help_dict:
            result += "\n\n\n%s.%s:\n\n%s" % (name, m, _help_dict[name + "." + m])
    return result


_help_dict = None


def get_docstring(objtype, symbol):
    """Get the docstring for object-type and symbol.

    Ex:
    get_docstring('Vector','sqrt')

    returns a string

    """
    global _help_dict
    if _help_dict is None:
        import neuron
        import os
        import zlib
        import pickle

        f = open(os.path.join(os.path.split(neuron.__file__)[0], "help_data.dat"), "rb")
        _help_dict = pickle.loads(zlib.decompress(f.read()))
        f.close()

    if (objtype, symbol) == ("", ""):

        return doc_h

    # are we asking for help on a class, e.g. h.Vector
    if objtype == "":
        if symbol in _help_dict:
            return _get_class_from_help_dict(symbol)
        else:
            return default_class_doc_template % symbol

    # are we asking for help on a object, e.g. h.Vector()
    if symbol == "":
        if objtype in _help_dict:
            return _get_class_from_help_dict(objtype)
        else:
            return default_object_doc_template % symbol

    # are we asking for help on a member of an object, e.g. h.Vector.size
    full_name = "%s.%s" % (objtype, symbol)
    if full_name in _help_dict:
        return _get_from_help_dict(full_name)
    else:

        return default_member_doc_template % (
            objtype,
            symbol,
            get_docstring(objtype, ""),
        )



def test_species_del(neuron_instance):
    """Test deleting an uninitialized species does not raise any error or
    exceptions.
    """

    h, rxd, data, save_path = neuron_instance
    soma = h.Section(name="soma")
    cyt = rxd.Region([soma])
    c = rxd.Species(cyt)
    try:
        c.__del__()
    except:
        assert False



def test_distance(neuron_instance):
    """Test 1D diffusion in a single section"""

    h, rxd, data, save_path = neuron_instance

    soma = h.Section(name="soma")
    dend = h.Section(name="dend")

    dend.connect(soma)

    cyt = rxd.Region([soma, dend], name="cyt")
    c = rxd.Species(cyt, name="c")

    assert h.distance(soma(0.5), dend(1)) == 150
    assert h.distance(soma(0.5), c.nodes[1]) == 100
    assert h.distance(c.nodes(soma), c.nodes(dend)) == 100

    h.distance(0, c.nodes(dend))
    assert h.distance(1, sec=dend) == 50

    assert h.distance(c.nodes[1], sec=soma) == 0
    assert h.distance(c.nodes(soma)) == 100



import os.path as osp
import numpy
import pytest
import gc

from .testutils import collect_data


def pytest_addoption(parser):
    parser.addoption("--mpi", action="store_true", default=False, help="use MPI")
    parser.addoption("--save", action="store", default=None, help="save the test data")


@pytest.fixture(scope="session")
def neuron_import(request):
    """Provides an instance of neuron h and rxd for tests"""

    # to use NEURON with MPI, mpi4py must be imported first.
    if request.config.getoption("--mpi"):
        from mpi4py import MPI  # noqa: F401

    save_path = request.config.getoption("--save")

    # we may not be not running in the test path so we have to load the mechanisms
    import neuron

    neuron.load_mechanisms(osp.abspath(osp.dirname(__file__)))
    from neuron import h, rxd

    return h, rxd, save_path


@pytest.fixture
def neuron_nosave_instance(neuron_import):
    """Sets/Resets the rxd test environment."""

    h, rxd, save_path = neuron_import
    h.load_file("stdrun.hoc")
    h.load_file("import3d.hoc")

    h.nrnunit_use_legacy(True)

    # pytest fixtures at the function scope that require neuron_instance will go
    # out of scope after neuron_instance. So species, sections, etc. will go
    # out of scope after neuron_instance is torn down.
    # Here we assert that no section left alive. If the assertion fails it is
    # due to a problem with the previous test, not the test which failed.
    gc.collect()
    for sec in h.allsec():
        assert False
    cvode = h.CVode()
    cvode.active(False)
    cvode.atol(1e-3)
    h.dt = 0.025
    h.stoprun = False

    yield (h, rxd, save_path)
    # With Python 3.11 on macOS there seemed to be problems related to garbage
    # collection happening during the following teardown
    gc.disable()
    for r in rxd.rxd._all_reactions[:]:
        if r():
            rxd.rxd._unregister_reaction(r)

    for s in rxd.species._all_species:
        if s():
            s().__del__()
    gc.enable()
    rxd.region._all_regions = []
    rxd.region._region_count = 0
    rxd.region._c_region_lookup = None
    rxd.species._species_counts = 0
    rxd.section1d._purge_cptrs()
    rxd.initializer.has_initialized = False
    rxd.rxd.free_conc_ptrs()
    rxd.rxd.free_curr_ptrs()
    rxd.rxd.rxd_include_node_flux1D(0, None, None, None)
    rxd.species._has_1d = False
    rxd.species._has_3d = False
    rxd.rxd._zero_volume_indices = numpy.ndarray(0, dtype=numpy.int_)
    rxd.set_solve_type(dimension=1)


@pytest.fixture
def neuron_instance(neuron_nosave_instance):
    """Sets/Resets the rxd test environment.
    Provides 'data', a dictionary used to store voltages and rxd node
    values for comparisons with the 'correct_data'.
    """

    h, rxd, save_path = neuron_nosave_instance
    data = {"record_count": 0, "data": []}

    def gather():
        return collect_data(h, rxd, data, save_path)

    cvode = h.CVode()
    cvode.extra_scatter_gather(0, gather)

    yield (h, rxd, data, save_path)

    cvode.extra_scatter_gather_remove(gather)



from neuron import h


class RxDSection:
    """parent class of all Section types"""

    def name(self):
        """The name of the section, as defined by HOC."""
        return self._sec.name()

    @property
    def region(self):
        """The Region containing the RxDSection."""
        return self._region

    @property
    def nseg(self):
        """The number of segments in the NEURON section.

        In some implementations, this might be different than the number of nodes in the section.
        """
        return self._nseg

    @property
    def nrn_region(self):
        """The HOC region, if any.

        'i' if the corresponding Region maps to HOC's internal concentration

        'o' if it maps to HOC's external concentration

        None otherwise.
        """

        return self._region.nrn_region

    @property
    def species(self):
        """The Species object stores on this RxDSection."""
        return self._species()

    @property
    def section_orientation(self):
        """The HOC section orientation."""
        return self._sec.orientation()

    @property
    def L(self):
        """The length of the section in microns."""
        return self._sec.L



# Basically want to test that FOR_NETCONS statement works when
# the NetCons connecting to ForNetConTest instances are created
# in random order.
import distutils.util
import os

from neuron import h

h.load_file("stdrun.hoc")

pc = h.ParallelContext()
h.dt = 1.0 / 32


class Cell:
    def __init__(self, gid):
        self.soma = h.Section(name="soma", cell=self)
        if gid % 2 == 0:
            # CoreNEURON permutation not the identity if cell topology not homogeneous
            self.dend = h.Section(name="dend", cell=self)
            self.dend.connect(self.soma(0.5))
        self.gid = gid
        pc.set_gid2node(gid, pc.id())
        self.r = h.Random()
        self.r.Random123(gid, 0, 0)
        self.syn = h.ForNetConTest(self.soma(0.5))
        pc.cell(gid, h.NetCon(self.syn, None))
        # random start times for the internal events
        self.syn.tbegin = self.r.discunif(0, 100) * h.dt


def test_fornetcon():
    from neuron import coreneuron

    coreneuron.enable = False

    ncell = 10
    gids = range(pc.id(), ncell, pc.nhost())  # round robin

    cells = [Cell(gid) for gid in gids]
    nclist = []

    # src first to more easily randomize NetCon creation order
    # so that NetCon to target not all adjacent
    for srcgid in range(ncell):
        for tarcell in cells:
            if int(tarcell.r.discunif(0, 1)) == 1:
                nclist.append(pc.gid_connect(srcgid, tarcell.syn))
                nclist[-1].delay = tarcell.r.discunif(10, 50) * h.dt
                nclist[-1].weight[0] = srcgid * 10000 + tarcell.gid

    spiketime = h.Vector()
    spikegid = h.Vector()
    pc.spike_record(-1, spiketime, spikegid)

    tstop = 8

    def run(tstop, mode):
        pc.set_maxstep(10)
        h.finitialize(-65)
        if mode == 0:
            pc.psolve(tstop)
        elif mode == 1:
            while h.t < tstop:
                pc.psolve(h.t + 1.0)
        else:
            while h.t < tstop:
                h.continuerun(h.t + 0.5)
                pc.psolve(h.t + 0.5)

    run(tstop, 0)  # NEURON run

    spiketime_std = spiketime.c()
    spikegid_std = spikegid.c()

    def get_weights():
        weight = []
        for nc in nclist:
            w = nc.weight
            weight.append((w[0], w[1], w[2], w[3], w[4]))
        return weight

    weight_std = get_weights()

    print("CoreNEURON run")
    h.CVode().cache_efficient(1)
    coreneuron.enable = True
    coreneuron.gpu = bool(
        distutils.util.strtobool(os.environ.get("CORENRN_ENABLE_GPU", "false"))
    )

    def runassert(mode):
        spiketime.resize(0)
        spikegid.resize(0)

        run(tstop, mode)
        assert len(spiketime) > 0
        assert spiketime_std.eq(spiketime) == 1.0
        assert spikegid_std.eq(spikegid) == 1.0
        assert len(weight_std) > 0
        assert weight_std == get_weights()

    for mode in [0, 1, 2]:
        runassert(mode)

    coreneuron.enable = False

    # help cover/test_netcvode.cpp cover the NetCon.setpost method
    # with respect to a change in weight vector size
    assert len(nclist[0].weight) == 5
    a = h.IntFire1()
    nclist[0].setpost(a)
    assert len(nclist[0].weight) == 1

    # teardown
    pc.gid_clear()


if __name__ == "__main__":
    test_fornetcon()
    h.quit()



from . import ctng
from . import scalarField
import numpy
from .surfaces import chunkify

_max_chunks = 10000000


def voxelize(
    source,
    dx=0.25,
    xlo=None,
    xhi=None,
    ylo=None,
    yhi=None,
    zlo=None,
    zhi=None,
    n_soma_step=100,
):
    """
    Generates a cartesian mesh of the volume of a neuron.

    Parameters
    ----------
    source : :func:`list`, ``nrn.SectionList``, or ``nrn.Import3D``
        The geometry to mesh.
    dx : double, optional
        Mesh step size.
    xlo : double, optional
        Minimum x value. If omitted or None, uses minimum x value in the geometry.
    xhi : double, optional
        Maximum x value. If omitted or None, uses maximum x value in the geometry.
    ylo : double, optional
        Minimum y value. If omitted or None, uses minimum y value in the geometry.
    yhi : double, optional
        Maximum y value. If omitted or None, uses maximum y value in the geometry.
    zlo : double, optional
        Minimum z value. If omitted or None, uses minimum z value in the geometry.
    zhi : double, optional
        Maximum z value. If omitted or None, uses maximum z value in the geometry.
    n_soma_step : integer, optional
        Number of pieces to slice a soma outline into.


    Returns
    -------
    result : :class:`ScalarField`
        The mesh. Values are scalars, but may be used as True inside the
        geometry and False outside.


    Examples
    --------

    Basic usage:

    >>> mesh = geometry3d.voxelize(h.allsec())

    Full example, using :mod:`pyplot`:

    >>> s1, s2, s3 = [h.Section() for i in xrange(3)]
    >>> for sec in [s2, s3]: ignore_return = sec.connect(s1)
    ...
    >>> for sec in h.allsec():
    ...     sec.diam = 1
    ...     sec.L = 5
    ...
    >>> mesh = geometry3d.voxelize(h.allsec(), dx=.1)
    >>> for i in xrange(10):
    ...     ignore_return = pyplot.subplot(2, 5, i + 1)
    ...     ignore_return = pyplot.imshow(mesh.values[:, :, i])
    ...     ignore_return = pyplot.xticks([])
    ...     ignore_return = pyplot.yticks([])
    ...
    >>> pyplot.show()

    .. plot::

        from neuron import h
        from matplotlib import pyplot
        from . import geometry3d

        s1, s2, s3 = [h.Section() for i in xrange(3)]
        for sec in [s2, s3]: ignore_return = sec.connect(s1)

        for sec in h.allsec():
            sec.diam = 1
            sec.L = 5

        mesh = geometry3d.voxelize(h.allsec(), dx=.1)
        for i in xrange(10):
            ignore_return = pyplot.subplot(2, 5, i + 1)
            ignore_return = pyplot.imshow(mesh.values[:, :, i])
            ignore_return = pyplot.xticks([])
            ignore_return = pyplot.yticks([])

        pyplot.show()



    .. note::
        The use of Import3D objects is recommended over lists of sections
        because the former preserves the soma outline information while
        the later does not. Up to one soma outline is currently supported.
    """

    objects = ctng.constructive_neuronal_geometry(source, n_soma_step, dx)

    if xlo is None:
        xlo = min(obj.xlo for obj in objects)
    if ylo is None:
        ylo = min(obj.ylo for obj in objects)
    if zlo is None:
        zlo = min(obj.zlo for obj in objects)
    if xhi is None:
        xhi = max(obj.xhi for obj in objects)
    if yhi is None:
        yhi = max(obj.yhi for obj in objects)
    if zhi is None:
        zhi = max(obj.zhi for obj in objects)

    mesh = scalarField.ScalarField(xlo, xhi, ylo, yhi, zlo, zhi, dx, dtype="B")
    grid = mesh.values

    # use chunks no smaller than 10 voxels across, but aim for max_chunks chunks
    chunk_size = max(
        10, int((len(mesh.xs) * len(mesh.ys) * len(mesh.zs) / _max_chunks) ** (1 / 3.0))
    )

    chunk_objs, nx, ny, nz = chunkify(
        objects, mesh.xs, mesh.ys, mesh.zs, chunk_size, dx
    )

    for i, x in enumerate(mesh.xs):
        chunk_objsa = chunk_objs[i // chunk_size]
        for j, y in enumerate(mesh.ys):
            chunk_objsb = chunk_objsa[j // chunk_size]
            for k, z in enumerate(mesh.zs):
                grid[i, j, k] = is_inside(x, y, z, chunk_objsb[k // chunk_size])

    return mesh


# inside the neuron if inside of any of its parts
def is_inside(x, y, z, active_objs):
    return 1 if any(obj(x, y, z) <= 0 for obj in active_objs) else 0
